// Copyright 2020 Kentaro Hibino. All rights reserved.
// Use of this source code is governed by a MIT license
// that can be found in the LICENSE file.

package asynq

import (
	"sync"
	"testing"
	"time"

	"github.com/google/go-cmp/cmp"
	"github.com/google/go-cmp/cmp/cmpopts"
	h "github.com/hibiken/asynq/internal/asynqtest"
	"github.com/hibiken/asynq/internal/base"
	"github.com/hibiken/asynq/internal/rdb"
)

// FIXME: Make this test better.
func TestHeartbeater(t *testing.T) {
	r := setup(t)
	rdbClient := rdb.NewRDB(r)

	tests := []struct {
		interval    time.Duration
		host        string
		pid         int
		queues      map[string]int
		concurrency int
	}{
		{time.Second, "some.address.ec2.aws.com", 45678, map[string]int{"default": 1}, 10},
	}

	timeCmpOpt := cmpopts.EquateApproxTime(10 * time.Millisecond)
	ignoreOpt := cmpopts.IgnoreUnexported(base.ProcessInfo{})
	for _, tc := range tests {
		h.FlushDB(t, r)

		stateCh := make(chan string)
		workerCh := make(chan int)
		hb := newHeartbeater(rdbClient, tc.host, tc.pid, tc.concurrency, tc.queues, false, tc.interval, stateCh, workerCh)

		want := &base.ProcessInfo{
			Host:        tc.host,
			PID:         tc.pid,
			Queues:      tc.queues,
			Concurrency: tc.concurrency,
			Started:     time.Now(),
			State:       "running",
		}
		var wg sync.WaitGroup
		hb.start(&wg)

		// allow for heartbeater to write to redis
		time.Sleep(tc.interval * 2)

		got, err := rdbClient.ReadProcessInfo(tc.host, tc.pid)
		if err != nil {
			t.Errorf("could not read process status from redis: %v", err)
			hb.terminate()
			continue
		}

		if diff := cmp.Diff(want, got, timeCmpOpt, ignoreOpt); diff != "" {
			t.Errorf("redis stored process status %+v, want %+v; (-want, +got)\n%s", got, want, diff)
			hb.terminate()
			continue
		}

		// state change
		stateCh <- "stopped"

		// allow for heartbeater to write to redis
		time.Sleep(tc.interval * 2)

		want.State = "stopped"
		got, err = rdbClient.ReadProcessInfo(tc.host, tc.pid)
		if err != nil {
			t.Errorf("could not read process status from redis: %v", err)
			hb.terminate()
			continue
		}

		if diff := cmp.Diff(want, got, timeCmpOpt, ignoreOpt); diff != "" {
			t.Errorf("redis stored process status %+v, want %+v; (-want, +got)\n%s", got, want, diff)
			hb.terminate()
			continue
		}

		hb.terminate()
	}
}
